############################### FUNDAMENTAL ###################################

package Fundamental;
@ISA = qw( _Buffer );

####################################
# Class Variables

# Emacs state
foreach $key (values %main::KeyCode){
	$InputStateKeyMap{$key} = \&InsertChar;
}
$InputStateKeyMap{$main::KeyCode{"DEL"}}=\&DeleteChar;
$InputStateKeyMap{$main::KeyCode{"CR"}}=\&SplitLine;
$InputStateKeyMap{$main::KeyCode{"KEY_BACKSPACE"}}=\&DeletePrevChar;
$InputStateKeyMap{$main::KeyCode{"KEY_UP"}}=\&MoveToPrevLine;
$InputStateKeyMap{$main::KeyCode{"KEY_DOWN"}}=\&MoveToNextLine;
$InputStateKeyMap{$main::KeyCode{"KEY_LEFT"}}=\&MoveToPrevChar;
$InputStateKeyMap{$main::KeyCode{"KEY_RIGHT"}}=\&MoveToNextChar;
$InputStateKeyMap{$main::KeyCode{"KEY_HOME"}}=\&MoveToLineBeginning;
$InputStateKeyMap{$main::KeyCode{"KEY_END"}}=\&MoveToLineEnd;
$InputStateKeyMap{$main::KeyCode{"^L"}}=\&RedrawDisplay;
$InputStateKeyMap{$main::KeyCode{"^A"}}=\&MoveToLineBeginning;
$InputStateKeyMap{$main::KeyCode{"^E"}}=\&MoveToLineEnd;
$InputStateKeyMap{$main::KeyCode{"^X"}}=\&GetMultiKey;
$InputStateKeyMap{$main::KeyCode{"ESC"}}=\&ChangeToCommandState;
$InputStateKeyMap{$main::KeyCode{"TAB"}}=\&Tab;
$InputStateKeyMap{$main::KeyCode{"^S"}}=\&IncrementalForwardSearch;
$InputStateKeyMap{$main::KeyCode{"^R"}}=\&IncrementalBackwardSearch;
$InputStateKeyMap{$main::KeyCode{"^V"}}=\&InsertNextKey;
$InputStateKeyMap{$main::KeyCode{"^Q"}}=\&InsertNextKey;

# Multiple Keystrokes Map
$MultiKeyMap{$main::KeyCode{"^F"}}=\&OpenFile;
$MultiKeyMap{$main::KeyCode{"^S"}}=\&SaveFile;
$MultiKeyMap{$main::KeyCode{"^W"}}=\&WriteFile;
$MultiKeyMap{$main::KeyCode{"^C"}}=\&Exit;
$MultiKeyMap{$main::KeyCode{"k"}}=\&Delete;
sub Delete { my $self = shift; $self->SUPER::Delete;}

# vi state
$CommandStateKeyMap{$main::KeyCode{"KEY_UP"}}=\&MoveToPrevLine;
$CommandStateKeyMap{$main::KeyCode{"KEY_DOWN"}}=\&MoveToNextLine;
$CommandStateKeyMap{$main::KeyCode{"KEY_LEFT"}}=\&MoveToPrevChar;
$CommandStateKeyMap{$main::KeyCode{"KEY_BACKSPACE"}}=\&MoveToPrevChar;
$CommandStateKeyMap{$main::KeyCode{"KEY_RIGHT"}}=\&MoveToNextChar;
$CommandStateKeyMap{$main::KeyCode{"l"}} = \&MoveToNextCharOnLine;
$CommandStateKeyMap{$main::KeyCode{"h"}} = \&MoveToPrevCharOnLine;
$CommandStateKeyMap{$main::KeyCode{"j"}} = \&MoveToNextLine;
$CommandStateKeyMap{$main::KeyCode{"k"}} = \&MoveToPrevLine;
$CommandStateKeyMap{$main::KeyCode{"^B"}} = \&MoveToPrevPage;
$CommandStateKeyMap{$main::KeyCode{"^F"}} = \&MoveToNextPage;
$CommandStateKeyMap{$main::KeyCode{"i"}} = \&ChangeToInputState;
$CommandStateKeyMap{$main::KeyCode{"a"}} = \&ChangeToInputStateAfter;
$CommandStateKeyMap{$main::KeyCode{"A"}} = \&ChangeToInputStateAfter;
$CommandStateKeyMap{$main::KeyCode{"J"}} = \&JoinLine;
$CommandStateKeyMap{$main::KeyCode{"^L"}} = \&RedrawDisplay;
$CommandStateKeyMap{$main::KeyCode{"x"}} = \&DeleteCharOnLine;
$CommandStateKeyMap{$main::KeyCode{"D"}} = \&DeleteToLineEnd;
$CommandStateKeyMap{$main::KeyCode{":"}} = \&GetLineCommand;
$CommandStateKeyMap{$main::KeyCode{"\$"}} = \&MoveToLineEnd;
$CommandStateKeyMap{$main::KeyCode{"0"}} = \&MoveToLineBeginning;
#$CommandStateKeyMap{$main::KeyCode{"0"}} = \&RepeatPrefix;
$CommandStateKeyMap{$main::KeyCode{"1"}} = \&RepeatPrefix;
$CommandStateKeyMap{$main::KeyCode{"2"}} = \&RepeatPrefix;
$CommandStateKeyMap{$main::KeyCode{"3"}} = \&RepeatPrefix;
$CommandStateKeyMap{$main::KeyCode{"4"}} = \&RepeatPrefix;
$CommandStateKeyMap{$main::KeyCode{"5"}} = \&RepeatPrefix;
$CommandStateKeyMap{$main::KeyCode{"6"}} = \&RepeatPrefix;
$CommandStateKeyMap{$main::KeyCode{"7"}} = \&RepeatPrefix;
$CommandStateKeyMap{$main::KeyCode{"8"}} = \&RepeatPrefix;
$CommandStateKeyMap{$main::KeyCode{"9"}} = \&RepeatPrefix;
$CommandStateKeyMap{$main::KeyCode{"G"}} = \&GotoRepeatPrefixLine;
$CommandStateKeyMap{$main::KeyCode{"/"}}=\&IncrementalForwardSearch;
$CommandStateKeyMap{$main::KeyCode{"?"}}=\&IncrementalBackwardSearch;
$CommandStateKeyMap{$main::KeyCode{"n"}}=\&FindNextTarget;
$CommandStateKeyMap{$main::KeyCode{"N"}}=\&FindPrevTarget;

####################################
# Instance Methods

# The _localize routine has the sole purpose of setting up the symbol
# tables so that the code is readable.  Putting $self-> in front of everything
# just made the code insufferable.  Anyway, now local methods and exported
# methods are clearly separated.
sub _localize {
	#_DebugLog("Fund _localize\n");
	my($sub) = shift;
        my($this) = shift;
	# Setup some convenient aliases
	local(*curcol)  = \$this->{'curcol'};
	local(*trycol)  = \$this->{"trycol"};
        local(*curline) = \$this->{'curline'};
	local(*buffer)  =  $this->{'buffer'};
	local(*curdisp) =  $this->{"curdisp"};
	local(*ro)      = \$this->{"read only"};
	local(*modified)= \$this->{"modified"};
	local(*curkeymap)= $this->{"curkeymap"};
	local(*self)    = \$this;
        return &$sub(@_);
}

sub Initialize {
	my($self) = shift;

	$self->SUPER::Initialize;

	#_DebugLog("Fund Initialize\n");
	$self->{"trycol"} = $self->{"curcol"} = 0;
	$self->{"curline"} = 0;
	$self->{"buffer"} = [];
	$self->{"ro"} = 0;
	$self->{"modified"} = 0;
	&_Fundamental($self);
	return $self;
}

# Call this to change modes on an existing buffer
sub FundamentalMode {
	my($self) = shift;

	bless $self; # change the inheritance
	&_Fundamental($self);
	return $self;
}

sub _Fundamental {
	my($self) = shift;
	$self->{"curdisp"} = []; 
        $self->{"mode desc"} = "Fundamental";
        $self->{"input state keymap"} = {%InputStateKeyMap};
        $self->{"command state keymap"} = {%CommandStateKeyMap};
        $self->ChangeToInputState;
}

sub GetLineCommand { return _localize(\&_GetLineCommand,@_); }
sub _GetLineCommand {
	my($command);
	$command = &_PromptStr(":");
	$self->DispatchLineCommand($command);
}

sub DispatchLineCommand {
	my($self) = shift;
	my($command) = shift;

	return if(!$command);
	if($command eq "w")	{ $self->SaveFile; return; } 
	if($command eq "wq")	{ $self->SaveFile;
				  $self->DeleteAndExitIfLast;
				  return; }
	if($command eq "q")	{ $self->DeleteAndExitIfLast; return; }
	if($command eq "q!")	{ $self->Exit; return; }
	if($command eq "n")	{ $self->SelectNextBuffer; return; }
	if($command eq "p")	{ $self->SelectPrevBuffer; return; }
	if($command eq "k")	{ $self->Delete; return; }
	if($command eq "help")	{ $self->ShowHelp; return; }

	my($prefix) = substr($command,0,1);
	if($prefix eq "%") {
		$self->Translate(substr($command,1));
		return;
	}

	eval { $self->$command(); };
	$self->{'curmsg'} = $@ if($@);
}

sub Translate {
	my($self) = shift;
	my($op) = shift;

	my(@newbuf) = split("\n",
		eval "my(\$res) = join('\n',\@buffer); \$res =~ $op; \$res;"
        );
	($self->{'curmsg'} = $@,return) if($@);
	@curdisp = undef;
	@buffer = @newbuf;
	&_SetModified;
}

sub RepeatPrefix { return _localize(\&_RepeatPrefix,@_); }
sub _RepeatPrefix {
	$$self{"RepeatPrefix"}.=chr($_[0]);
}

sub GetMultiKey { return _localize(\&_GetMultiKey,@_); }
sub _GetMultiKey {
	&_ShowMsg("^X");
	my($key)=&_GetKey;
	&_DispatchKey($key,\%MultiKeyMap);
}

sub _ShowHelp { $self->ShowHelp; }
sub ShowHelp {
	my($self) = shift;
	#&_DebugLog("Fundamental: ShowHelp\n");
	my($newbuf) = new Fundamental;
	$$newbuf{"read only"} = 1;
	my($newview) = new View ("full",$newbuf);
	$newview->BecomeEditView;
	($$newbuf{"buffername"},$$newbuf{"buffer"}) = $self->Help;
}

sub Help {
	return "Fundamental HELP",[ split "\n", <<HELP_TEXT

                         ped Fundamental Mode Help

   ped is this cool editor I wrote using perl.  I like both vi and emacs for
editing and I like perl for text manipulation.  Therfore, ped is written in
perl so I can modify it easily and it functions a little like vi and emacs.
Plus, since I wrote it, I know just how it works! :)

Send patches to me, Matt Agler, at matagl\@sypher.com
   Please use unified diff format if possible.

The status line displays INPUT when in emacs mode and COMMAND when in vi mode.

To enter COMMAND mode from INPUT mode, hit ESC.  To enter INPUT mode from
COMMAND mode, hit i.

   INPUT    COMMAND  Description
   -------- -------- ----------------------------------------------------------
Change Modes
   ESC               Change to COMMAND mode
            i        Change to INPUT mode
            a/A      Change to INPUT mode after current character
            :HTML::HTMLMode Switch to HTML Mode
            :XML::XMLMode   Switch to XML Mode

Movement
   UP       k/UP     Move to previous line
   DOWN     j/DOWN   Move to next line
   LEFT     LEFT/BS  Move to previous character
   RIGHT    RIGHT    Move to next character
            h        Move to previous character on line
            l        Move to next character on line
   ^A/HOME  0        Move to beginning of line
   ^E/END   \$        Move to end of line
            ^B       Move to previous page
            ^F       Move to next page
            <N> G    Move to line <N>

File Handling
   ^X^F              Open new file for editing
   ^X^S     :w       Save file
            :wq      Save file and end editing
   ^X^W              Write file with new filename
            :q       End editing
   ^X^C     :q!      End editing and Exit
            :k       Delete the current buffer
            :help    Show help

Editing
   DEL      x        Delete the current character
            J        Join lines
            D        Delete to end of line
   ^L                Refresh the display
   CR                Split the current line
   BS                Delete the previous character
   ^Q/^V             Insert next keypress

Search
   ^S       /        Begin forward incremental search
   ^S^S     n        Find next match for current search
   ^R       ?        Begin backward incremental search
   ^R^R     N        Find prev match for current search
            [*]      Search forward for identifier under cursor  [Not Implemented]
            [#]      Search backward for identifier under cursor [Not Implemented]

Search and Replace
            :%       The entire buffer is converted to a single string and
                     passed to the command.  For example:
            :%s/This/That/gm  Search and replace
            :%tr[a-z][A-Z]    Translate lowercase to uppercase
            :%tr/^M//d        Delete all ^Ms

HELP_TEXT
];
}

sub EditViewChanged {
	my($self) = shift;

	$self->{'curcol'} = shift;
	$self->{'curline'} = shift;

	# we never want to return to a search in progress
	$self->{'incrementalSearch'} = 0;
}

sub DispatchKey { return _localize(\&_DispatchKey,@_); }
sub _DispatchKey {
	my($key,$map) = @_;
	#&_DebugLog("Fund _KeyDispatch\n");
	$map = \%curkeymap if(!defined $map);

	# Check if we're in mid incremental search
	if($incrementalSearch){
		if($key == $main::KeyCode{"^S"} and
		   $self->{"state"} eq "INPUT"){
			&_FindNextTarget;
		} elsif($key == $main::KeyCode{"^R"} and
		        $self->{"state"} eq "INPUT"){
			&_FindPrevTarget;
		} elsif($key == $main::KeyCode{"KEY_UP"} ||
		   $key == $main::KeyCode{"KEY_DOWN"} ||
		   $key == $main::KeyCode{"KEY_LEFT"} ||
		   $key == $main::KeyCode{"KEY_RIGHT"} ||
		   $key == $main::KeyCode{"CR"}){
			&_EndIncrementalSearch;
			return;
		} else {
			&_RefineIncrementalSearch($key);
			return;
		}
	}
	$self->SUPER::DispatchKey($key,$map);
}

sub DispXY { return _localize(\&_DispXY,@_); }
sub _DispXY {
	#&_DebugLog("Fund _DispXY\n");
	# locate ourselves on the current line
	my($lhalf,$str,$dispcol);
	if($self->{'state'} eq "INPUT"){
		# we position on the start of wide chars
		$str = $buffer[$curline];
		if(&_LineEnd){
			$lhalf = &_Expand($str);
		} else {
			$lhalf = &_Expand(substr($str,0,$curcol));
		}
		$dispcol = length($lhalf);
	} else {
		# we position on the end of wide chars
		$str = $buffer[$curline];
		$curcol++; #we do this so that LineEnd works
		if(&_LineEnd){
			$lhalf = &_Expand($str);
		} else {
			$lhalf = &_Expand(substr($str,0,$curcol));
		}
		$curcol--;
		$dispcol = length($lhalf) - (&_LineEnd?0:1);
	}
	return ($dispcol,$curline);
}
	
sub Display { return _localize(\&_Display,@_); }
sub _Display {
	my($curleft,$curtop,$wid,$hig) = @_;

	my(@display,$i);

	# Generate the display
	for($i=$curtop;$i < $curtop+$hig; $i++){
		$curdisp[$i] = &_Expand($buffer[$i]) if(!defined $curdisp[$i]);
		my($len) = length($curdisp[$i]);
		$display[$i-$curtop] =
			($len < $curleft)?"":substr($curdisp[$i],$curleft,$wid);
	}

        my($statline);
        $statline  = "=";
        $statline .= $$self{"read only"}?"RO":"";
        $statline .= $$self{"modified"}?"##":"";
        $statline .= "="x(4-length($statline));

        $statline .= $$self{"buffername"};
        $statline .= "="x($wid-length($statline)-35);

        $statline .= $$self{"mode desc"};
        $statline .= "="x($wid-length($statline)-20);

        $statline .= "C".(1+$$self{"curcol"});
        $statline .= "="x($wid-length($statline)-14);
        $statline .= "L".(1+$$self{"curline"});
        $statline .= "="x($wid-length($statline)-8);

        $statline .= $$self{"state"};
        $statline .= "="x($wid-length($statline));

	return($statline,@display);
}

sub Expand { return _localize(\&_Expand,@_); }
sub _Expand {
	my($str) = (@_);

	#expand tabs
	$str =~ s/\011/   /g;

	# put a carat before control chars
	$str =~ s/([\000-\037])/\^$1/g;
	$str =~ tr[\000-\037][\100-\137]; # make control chars viewable

	return($str);
}

###################
# Change States

sub _ChangeToCommandState { $self->ChangeToCommandState; }
sub ChangeToCommandState {
	my($self) = shift;
	$self->MoveToPrevCharOnLine;
	$self->{'curkeymap'} = $self->{"command state keymap"};
	$self->{'state'} = "COMMAND";
}

sub _ChangeToInputState { $self->ChangeToInputState; }
sub ChangeToInputState {
	my($self) = shift;
	$self->{'curkeymap'} = $self->{"input state keymap"};
	$self->{"state"} = "INPUT";
}

sub ChangeToInputStateAfter { return _localize(\&_ChangeToInputStateAfter,@_); }
sub _ChangeToInputStateAfter {
	&_MoveToNextCharOnLine;
	$self->ChangeToInputState;
}

###################
# Search and Replace

sub SearchAndReplace { return _localize(\&_LookingAt,@_); }
sub _SearchAndReplace {
	my($re,$replacement) = @_;

	map s/$re/$replacement/, @buffer;
}

###################
# Incremental Searching

sub _ShowSearch { $self->{"curmsg"} = "ISearch: " . $$self{"searchTarget"}; }

sub EndIncrementalSearch { return _localize(\&_EndIncrementalSearch,@_); }
sub _EndIncrementalSearch {
	#&_DebugLog("EndIncrementalSearch\n");
	$incrementalSearch = 0;
	$$self{"curmsg"} = "";
}

sub RefineIncrementalSearch { return _localize(\&_RefineIncrementalSearch,@_); }
sub _RefineIncrementalSearch {
	my($key) = @_;
	my($char) = chr($key);

	$$self{"oldTarget"} = $$self{"searchTarget"} .= $char;
	#&_DebugLog("RefineIncrementalSearch to ".$$self{"searchTarget"}."\n");
	if(!&_LookingAt($$self{"searchTarget"})){
		if($incrementalSearchDir) {
			# Forward
			&_FindNextTarget;
		} else {
			&_FindPrevTarget;
		}
	}
	&_ShowSearch;
}

sub LookingAt { return _localize(\&_LookingAt,@_); }
sub _LookingAt {
	my($target) = @_;
	return substr($buffer[$curline],$curcol) =~ m/^$target/;
}

sub FindNextTarget { return _localize(\&_FindNextTarget,@_); }
sub _FindNextTarget {
	return &_FindNextRE($$self{"searchTarget"});
}

sub FindNextRE { return _localize(\&_FindNextRE,@_); }
sub _FindNextRE {
	my($re)=shift;
	my($line);
	foreach (&_LineEnd?"":substr($buffer[$curline],$curcol+1),
		 @buffer[$curline+1..$#buffer])
	{
		if(m/$re/){
			$curline = $curline + $line;
			$trycol = $curcol = length($`)
				+ ($line ? 0 : $curcol+1);
			return 1;
		}
		$line++;
	}
	return 0;
}

sub IncrementalForwardSearch { return _localize(\&_IncrementalForwardSearch,@_); }
sub _IncrementalForwardSearch {
	$incrementalSearchDir = 1; #Forwards
	if($incrementalSearch){
		if($$self{"searchTarget"} eq "(?i)"){
			$$self{"searchTarget"} = $$self{"oldTarget"};
		}
		&_FindNextTarget;
	} else {
		$incrementalSearch = 1;
		$$self{"searchTarget"} = "(?i)";
	}
	&_ShowSearch;
}

sub FindPrevTarget { return _localize(\&_FindPrevTarget,@_); }
sub _FindPrevTarget {
	return &_FindPrevRE($$self{"searchTarget"});
}

sub FindPrevRE  { return _localize(\&_FindPrevRE ,@_); }
sub _FindPrevRE {
	my($re)=shift;
	my($line,$col,$target);
	foreach (reverse(@buffer[0..$curline]))
	{
		if(m/$re/ &&
		   ($line || length($`) < ($curcol-1))
		  )
		{
			$col = length($`);
			$target = substr($_,length($`)+1);
			while($target =~ m/$re/){
				last if($line == 0 &&
					($col + length($`) + 1) >= ($curcol-1));
				$col += length($`) + 1;
				$target = substr($target,length($`)+1);
			}
			$trycol = $curcol = $col;
			$curline = $curline - $line;
			return 1;
		}
		$line++;
	}
	return 0;
}

sub IncrementalBackwardSearch { return _localize(\&_IncrementalBackwardSearch,@_); }
sub _IncrementalBackwardSearch {
	$incrementalSearchDir = 0; #Backwards
	if($incrementalSearch){
		if($$self{"searchTarget"} eq "(?i)"){
			$$self{"searchTarget"} = $$self{"oldTarget"};
		}
		&_FindPrevTarget;
	} else {
		$incrementalSearch = 1;
		$$self{"searchTarget"} = "(?i)";
	}
	&_ShowSearch;
}

###################
#Location Checking

sub BufferBeginning { return _localize(\&_BufferBeginning,@_); }
sub _BufferBeginning { return !($curcol || $curline); }

sub BufferEnd { return _localize(\&_BufferEnd,@_); }
sub _BufferEnd       {
	return ($curline >= $#buffer
		&& $curcol >= length($buffer[$curline]));
}

sub LineBeginning { return _localize(\&_LineBeginning,@_); }
sub _LineBeginning   { return !$curcol; }

sub LineEnd { return _localize(\&_LineEnd,@_); }
sub _LineEnd         { return ($curcol >= length($buffer[$curline])); }

###################
#Movement

sub _TryForCol {
	$curcol = $trycol;
	if(length($buffer[$curline])<$curcol){
		$curcol = length($buffer[$curline]);
	}
}

sub GotoRepeatPrefixLine { return _localize(\&_GotoRepeatPrefixLine,@_); }
sub _GotoRepeatPrefixLine {
	($curline = $#buffer, return) if(!defined($$self{"RepeatPrefix"}));
	$curline = $$self{"RepeatPrefix"} - 1;
	$$self{"RepeatPrefix"} = undef;
	$curline = 0 if($curline<0);
	$curline = $#buffer if($curline>$#buffer);
	&_TryForCol;
}

sub MoveToXY { return _localize(\&_MoveToXY,@_); }
sub _MoveToXY {
	$trycol = $curcol = shift;
	$curline = shift;
	#_DebugLog("Fund _MoveToXY $curcol $curline\n");
	$curline = $#buffer if($curline > $#buffer);
	$curline = 0 if($curline < $#buffer || $#buffer < 0);
	$trycol = $curcol = 0 if($curcol < 0);
	$trycol = $curcol = 
	   length($buffer[$curline]) if(length($buffer[$curline]) < $curcol);
}

sub MoveToBufferEnd { return _localize(\&_MoveToBufferEnd,@_); }
sub _MoveToBufferEnd {
	$curline = $#buffer;
	$curcol = length($buffer[$#buffer]);
}

sub MoveToPrevPage { return _localize(\&_MoveToPrevPage,@_); }
sub _MoveToPrevPage {
	$curline -= $$self{"LINES"};
	$curline = 0 if($curline<0);
	$curtop -= $$self{"LINES"};
	$curtop = 0 if($curtop<0);
	&_TryForCol;
}

sub MoveToNextPage { return _localize(\&_MoveToNextPage,@_); }
sub _MoveToNextPage {
	$curline += $$self{"LINES"};
	$curline = $#buffer if($curline>$#buffer);
	$curtop += $$self{"LINES"};
	$curtop = $#buffer if($curtop>$#buffer);
	&_TryForCol;
}
sub MoveToPrevLine { return _localize(\&_MoveToPrevLine,@_); }
sub _MoveToPrevLine {
	$curline || return;
	$curline--;
	&_TryForCol;
}

sub MoveToNextLine { return _localize(\&_MoveToNextLine,@_); }
sub _MoveToNextLine {
	($curline < $#buffer) || return;
	$curline++;
	&_TryForCol;
}

sub MoveToPrevChar { return _localize(\&_MoveToPrevChar,@_); }
sub _MoveToPrevChar {
	#$$self{"curmsg"} = "MoveToPrevChar";
	&_BufferBeginning && return;
	($trycol = --$curcol,return) if(!&_LineBeginning);
	$curline--;
	$trycol = $curcol=length($buffer[$curline]);
}

sub MoveToNextChar { return _localize(\&_MoveToNextChar,@_); }
sub _MoveToNextChar {
	&_BufferEnd && return;
	($trycol = ++$curcol,return) if(!&_LineEnd);
	$curline++;
	$trycol = $curcol = 0;
}

sub MoveToNextCharOnLine { return _localize(\&_MoveToNextCharOnLine,@_); }
sub _MoveToNextCharOnLine {
	&_MoveToNextChar if(!&_LineEnd);
}

sub MoveToPrevCharOnLine { return _localize(\&_MoveToPrevCharOnLine,@_); }
sub _MoveToPrevCharOnLine {
	&_MoveToPrevChar if(!&_LineBeginning);
}

sub MoveToLineBeginning { return _localize(\&_MoveToLineBeginning,@_); }
sub _MoveToLineBeginning {
	$trycol = $curcol = 0;
}

sub MoveToLineEnd { return _localize(\&_MoveToLineEnd,@_); }
sub _MoveToLineEnd { $trycol = $curcol = length($buffer[$curline]); }

###################
#Line Manipulation

sub JoinLine { return _localize(\&_JoinLine,@_); }
sub _JoinLine {
	$ro && return;
	return if($#buffer <= $curline);
	@curdisp[$curline..$#curdisp] = undef;
	$buffer[$curline] .= $buffer[$curline + 1];
	splice(@buffer,$curline + 1,1);
	&_SetModified;
}

sub SplitLine { return _localize(\&_SplitLine,@_); }
sub _SplitLine {
	$ro && return;
	@curdisp[$curline..$#curdisp] = undef;
	splice(@buffer,$curline + 1,0,(""));
	$buffer[$curline + 1] = substr($buffer[$curline],$curcol);
	$buffer[$curline] =
		&_LineBeginning?"":substr($buffer[$curline],0,$curcol);
	&_MoveToLineBeginning;
	&_MoveToNextLine;
	&_SetModified;
}

sub DeleteToLineEnd { return _localize(\&_DeleteToLineEnd,@_); }
sub _DeleteToLineEnd {
	$ro && return;
	my($buffer) = $$self{"buffer"};
	my($curline) = $$self{"curline"};
	my($curcol) = $$self{"curcol"};

	$curdisp[$curline]=undef;
	$buffer[$curline]=substr($buffer[$curline],0,$curcol);
	&_SetModified;
}

sub InsertString { return _localize(\&_InsertString,@_); }
sub _InsertString {
	$ro && return;
	my($str) = shift;
	$curdisp[$curline]=undef;
	if(&_LineEnd){
                $buffer[$curline] .=  $str;
        } else {
                $buffer[$curline] =
                        substr($buffer[$curline],0,$curcol)
                        . $str
                        . substr($buffer[$curline],$curcol);
        }
        $trycol = $curcol += length($str);
        &_SetModified;
}

###################
#Character Manipulation

sub DeleteCharOnLine { return _localize(\&_DeleteCharOnLine,@_); }
sub _DeleteCharOnLine {
	$ro && return;
	&_DeleteChar if(!&_LineEnd);
}

sub DeleteChar { return _localize(\&_DeleteChar,@_); }
sub _DeleteChar {
	$ro && return;
	&_BufferEnd && return;
	(&_JoinLine,return) if($curcol+1 > length($buffer[$curline]));
	$curdisp[$curline]=undef;
	$buffer[$curline] =
		substr($buffer[$curline],0,$curcol)
		. substr($buffer[$curline],$curcol+1);
	&_SetModified;
}

sub DeletePrevChar { return _localize(\&_DeletePrevChar,@_); }
sub _DeletePrevChar {
	$ro && return;
	&_BufferBeginning && return;
	&_MoveToPrevChar;
	&_DeleteChar;
}
sub DeletePrevCharOnLine { return _localize(\&_DeletePrevCharOnLine,@_); }
sub _DeletePrevCharOnLine {
	$ro && return;
	return if(&_LineBeginning);
	&_MoveToPrevChar;
	&_DeleteChar;
}
sub InsertChar { return _localize(\&_InsertChar,@_); }
sub _InsertChar { &_InsertString(chr($_[0])); }

sub InsertNextKey {
	my($key)=&_GetKey;
	&_InsertChar($key);
}

sub Tab { return _localize(\&_Tab,@_); }
sub _Tab { &_InsertString("   "); }

###################
# Buffer Manipulation
# and Interface to main::

sub SetModified { return _localize(\&_SetModified,@_); }
sub _SetModified { $modified = 1; }

sub SetNotModified { return _localize(\&_SetNotModified,@_); }
sub _SetNotModified { $modified = 0; }

sub Exit { return _localize(\&_Exit,@_); }
sub _Exit { &main::ExitOnReturn; }

sub RedrawDisplay { return _localize(\&_RedrawDisplay,@_); }
sub _RedrawDisplay { @curdisp = undef; View->ForceRedraw; }

sub ChooseFile {
        #&DebugLog("ped PromptFile\n");
	my($self) = shift;
        my($buf) = new MiniBuffer @_;
        my($newview) = new View ("mini",$buf);
        $newview->BecomeEditView;
}  

sub OpenFile { return _localize(\&_OpenFile,@_); }
sub _OpenFile {
	#&_DebugLog("Fund _OpenFile\n");
	my($closure) = sub {
		my($filename) = shift;
		my($newbuf)= new Fundamental;
		$$newbuf{"filename"}=$filename;
		$$newbuf{"buffername"}=$filename;
		$newbuf->ReReadFile;
		my($newview) = new View ("full",$newbuf);
		$newview->BecomeEditView;
	};
	#&_DebugLog("Fund OpenFile passing closure ".$closure."\n");
	$self->ChooseFile("Open File: ",$ENV{HOME},$closure);
}

sub WriteFile { return _localize(\&_WriteFile,@_); }
sub _WriteFile {
	my($this) = $self;	
	my($closure) = sub {
                my($filename) = shift;
                $$this{"filename"}=$filename;
                $$this{"buffername"}=$filename;
		$this->SaveFile;
        };
        $self->ChooseFile("Write file: ",$self->{"filename"},$closure);
}

sub ShowMsg { return _localize(\&_ShowMsg,@_); }
sub _ShowMsg { View->ShowMsg(@_); }

sub GetKey { return _localize(\&_GetKey,@_); }
sub _GetKey { return View->GetKey(@_); }

sub PromptStr { return _localize(\&_PromptStr,@_); }
sub _PromptStr { return View->PromptStr(@_); }

sub DebugLog { return _localize(\&_DebugLog,@_); }
sub _DebugLog { &main::DebugLog(@_); }

sub ReReadFile { return _localize(\&_ReReadFile,@_); }
sub _ReReadFile {
        my($lastchar,@file);

        open FILE,$self->{'filename'};
        @file = <FILE>;
        close FILE;
        if($#file < 0){ $self->{'buffer'} = []; return;}
        $lastchar = chop @file;
        if($lastchar ne "\n"){
                $file[$#file] .= $lastchar;
                $self->{"no trailing newline"} = 1;
        }
        $self->{'buffer'} = \@file;
	&_SetNotModified;
	@curdisp = undef;
}

sub SaveFile { return _localize(\&_SaveFile,@_); }
sub _SaveFile {
        if(!$self->{'filename'}){
                return $self->WriteFile;
        }
        open(FILE, ">$self->{'filename'}") or do {
                $self->{"curmsg"} = "Unable to open: " . $self->{'filename'};
                return 0;
        };
        print FILE join("\n",@{$self->{"buffer"}}),
                ($self->{"no trailing newline"}?"":"\n");
        close FILE;
	&_SetNotModified;
        return 1;
}

1;
